/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * SCIDYNE Corporation, 649 School Street, Pembroke MA 02359 USA
 * File: multi-io-ard_Adafruit_library_demo.ino
 * Compiler: Arduino IDE Ver 1.8.10
 * Date: August 2, 2023
 * Author: Mark Durgin 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * 
 * This sketch demonstrates the basic steps necessary to setup and begin using 
 * the SCIDYNE MULTI-IO-ARD (PN #100-7714) with Adafruit open-source libraries. 
 * Please refer to the actual library files and other documentation for
 * more thorough library usage information.
 * 
 * SCIDYNE is not affiliated with Adafruit but their contributions and products 
 * are appreciated. Please support Adafruit by visiting www.adafruit.com
 * 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * The program continuously generates digital pulses on certain Pins as defined within the software. 
 * Analog outputs are set to predefined values. Analog inputs are scanned and the results displayed
 * as RAW values and also expressed as voltages.  Program operation can be observed using
 * the Arduino IDE Serial Monitor, an oscilloscope, and voltage meter. 
 * 
 * For convenience, the eight Analog Output channels can be fed back into the eight Analog Input channels.
 * Connect AOCH0 to AICH0, AOCH1 to AICH1, and so on.
 * 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * The Adafruit libraries supporting the MCP23017, MCP4728, and ADS1X15 chips 
 * must be available within the Arduino IDE. If not installed already, go to  
 * Sketch > Include Library > Manage Libraries.  Search for the three chips 
 * when prompted and follow the installation instructions.
 * 
 * Alternatively, the Adafruit libraries can be downloaded directly from the 
 * Adafruit github repository and installed manually:
 * 
 * Digital I/O,   MCP23017:  https://github.com/adafruit/Adafruit-MCP23017-Arduino-Library
 * Analog Outputs, MCP4728:  https://github.com/adafruit/Adafruit_MCP4728
 * Analog Inputs,  ADS1115:  https://github.com/adafruit/Adafruit_ADS1X15
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  
 *  _   ___   __  ___  ___                 ___       _  
 * | \   |   /__   |    |    /\   |         |    /  / \ 
 * |_/  _|_  \_|  _|_   |   /--\  |_       _|_  /   \_/ 
 *                                                  
 * Summary of the MCP23017 Digital I/O library routines:
 * // Initialization
 * bool begin_I2C( uint8_t i2c_addr );
 *
 * // Main Arduino API methods
 * void pinMode( uint8_t pin, uint8_t mode );
 * uint8_t digitalRead( uint8_t pin );
 * void digitalWrite( uint8_t pin, uint8_t value );
 *
 * // Bulk access
 * uint8_t readGPIO( uint8_t port );
 * void writeGPIO( uint8_t value, uint8_t port );
 *
 * // Interrupts
 * void setupInterrupts( bool mirroring, bool openDrain, uint8_t polarity );
 * void setupInterruptPin( uint8_t pin, uint8_t mode );
 * void disableInterruptPin( uint8_t pin );
 * uint8_t getLastInterruptPin();
 * 
 * Library Usage Hints:
 *  - Values for Pin parameters are 0 - 15 corresponding to the 8-bits of a PortA (Pins 0 - 7) and 
 *    the 8-Bits of a PortB (Pins 8 - 15).
 *  - Values for Port parameters are 0 for PortA and 1 for PortB
 * 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 *                        _    __        _        ___   _        ___   __ 
 *  /\   |\ |   /\   |   / \  /__       / \  | |   |   |_)  | |   |   (_  
 * /--\  | \|  /--\  |_  \_/  \_|       \_/  |_|   |   |    |_|   |   __)  
 * 
 * Summary of the MCP4728 Analog Output library routines: 
 * bool begin( uint8_t i2c_address );
 *
 * bool setChannelValue( MCP4728_channel_t channel,
 *                       uint16_t new_value,
 *                       MCP4728_vref_t new_vref = MCP4728_VREF_VDD,
 *                       MCP4728_gain_t new_gain = MCP4728_GAIN_1X,
 *                       MCP4728_pd_mode_t new_pd_mode = MCP4728_PD_MODE_NORMAL,
 *                       bool udac = false);
 * bool fastWrite( uint16_t channel_a_value, uint16_t channel_b_value,
 *                 uint16_t channel_c_value, uint16_t channel_d_value);
 *                
 * bool saveToEEPROM( void ); 
 * 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *                                                                                    
 *                        _    __       ___         _        ___   __ 
 *  /\   |\ |   /\   |   / \  /__        |   |\ |  |_)  | |   |   (_  
 * /--\  | \|  /--\  |_  \_/  \_|       _|_  | \|  |    |_|   |   __) 
 *                                                                                                                                                      
 * Summary of the ADS1X15 Analog Input library routines:
 * bool begin( uint8_t i2c_addr );
 * int16_t readADC_SingleEnded( uint8_t channel );
 * int16_t readADC_Differential_0_1();
 * int16_t readADC_Differential_2_3();
 * void startComparator_SingleEnded(uint8_t channel, int16_t threshold);
 * int16_t getLastConversionResults();
 * float computeVolts( int16_t counts );
 * void setGain( adsGain_t gain );
 * adsGain_t getGain();
 * void setDataRate( uint16_t rate );
 * uint16_t getDataRate();
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * 
 */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Include necessary files
 */
#include <Wire.h>                 // Needed for I2C communications
#include <Adafruit_MCP23X17.h>    // Include the Adafruit MCP23017 library, supports the Digital I/O
#include <Adafruit_MCP4728.h>     // Include the Adafruit MCP4728 library, supports the Analog Outputs
#include <Adafruit_ADS1X15.h>     // Include the Adafruit ADS1X15 library, supports the Analog Inputs

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Define the I2C addresses of the chips used on the MULTI-IO-ARD.
 * Make sure no other I2C devices use these addresses that would cause a conflict.
 */
#define MCP23017_ADDR   0x20    // I2C Address of the MCP23017 Digital I/O chip (U1)
#define MCP4728_A_ADDR  0x60    // I2C Address of the MCP4728 DAC chip (U2)controlling AOCH[3:0]
#define MCP4728_B_ADDR  0x61    // I2C Address of the MCP4728 DAC chip (U3)controlling AOCH[7:4]
#define ADS1115_A_ADDR  0x48    // I2C Address of the ADS1115 ADC chip (U4)controlling AICH[3:0]
#define ADS1115_B_ADDR  0x49    // I2C Address of the ADS1115 ADC chip (U5)controlling AICH[7:4]


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Define program constants
 */
#define PORT_PIN     0   // The MCP23017 Port Pin to toggle. Valid number is 0-15, where 0 = PortA.0 ... 15 = PortB.7
#define PORT_PULSCNT 3   // Number of pulses to generate on the designated PortA Pin

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Create instances of the Adafruit libraries
 */
Adafruit_MCP23X17 MCP23017;    // Create instance of library to control the MCP23017 digital I/O chip U1
Adafruit_MCP4728  MCP4728_A;   // Create instance of library to control the MCP4728 Analog Output DAC chip U2
Adafruit_MCP4728  MCP4728_B;   // Create instance of library to control the MCP4728 Analog Output DAC chip U3
Adafruit_ADS1115  ADS1115_A;   // Create instance of library to control the ADS1115 Analog Input ADC chip U4
Adafruit_ADS1115  ADS1115_B;   // Create instance of library to control the ADS1115 Analog Input ADC chip U5


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * setup runs once then execution goes to loop
 */
void setup()
{
  Wire.begin();                           // Setup I2C communications

  Serial.begin(9600);                     // Setup communications to IDE Monitor
    while (!Serial);
  
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * Initialize the library associated with the digital I/O MCP23017 chip PortA and PortB
  */ 
  if (!MCP23017.begin_I2C(MCP23017_ADDR)) {
    Serial.println(F("Error initializing MCP23017 U1"));
    while (1);
  }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * Configure the designated pins on the MCP23017 PortA
   * Where mode can be:
   *  OUTPUT            Configures Pin as an output 
   *  INPUT             Configures Pin as Input
   *  INPUT_PULLUP      Configures Pin as input with Pull-Up enabled
   */
  MCP23017.pinMode(PORT_PIN, OUTPUT);  // This is the pin that will generate the pulses


  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * Initialize the library associated with Analog Outputs AOCH[3:0], MCP4728 Quad-DAC chip U2, 
  */ 
  if (!MCP4728_A.begin(MCP4728_A_ADDR)) {
    Serial.println(F("Error initializing MCP4728_A U2"));
    while (1);
    }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * Initialize the library associated with Analog Outputs AOCH[7:4], MCP4728 Quad-DAC chip U3, 
  */ 
  if (!MCP4728_B.begin(MCP4728_B_ADDR)) {
    Serial.println(F("Error initializing MCP4728_A U3"));
    while (1);
    }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * Initialize the library associated with the Analog Inputs AICH[3:0], ADS1115 ADC chip U4
  */ 
  if (!ADS1115_A.begin(ADS1115_A_ADDR)) {
    Serial.println(F("Error initializing ADS1115_A U4"));
    while (1);
  }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * Initialize the library associated with the Analog Inputs AICH[7:4], ADS1115 ADC chip U5 
  */ 
  if (!ADS1115_B.begin(ADS1115_B_ADDR)) {
    Serial.println(F("Error initializing ADS1115_B U5"));
    while (1);
  }

  /* * * * * * * * * * * * * * * * * * * * * * * 
   * The ADS1115 ADC input range (or gain) can be changed by the these 
   * .setGain() function. Be careful never to exceed VDD +0.3V max, or to
   * exceed the upper and lower limits. Appling voltages exceeding the 
   * upper or lower limits may permanently damage the MULTI-IO-ARD!
   *                                Gain         Range      Resolution
   * .setGain(GAIN_TWOTHIRDS);  // 2/3x gain +/- 6.144V  1 bit = 0.1875mV (default)
   * .setGain(GAIN_ONE);        // 1x gain   +/- 4.096V  1 bit = 0.125mV
   * .setGain(GAIN_TWO);        // 2x gain   +/- 2.048V  1 bit = 0.0625mV
   * .setGain(GAIN_FOUR);       // 4x gain   +/- 1.024V  1 bit = 0.03125mV
   * .setGain(GAIN_EIGHT);      // 8x gain   +/- 0.512V  1 bit = 0.015625mV
   * .setGain(GAIN_SIXTEEN);    // 16x gain  +/- 0.256V  1 bit = 0.0078125mV
   */
   ADS1115_A.setGain(GAIN_TWOTHIRDS);
   ADS1115_B.setGain(GAIN_TWOTHIRDS);

   Serial.println(F("Looping..."));  // Tell user program is running

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * loop runs continuously
 */
void loop()
{
  uint8_t i;                                                              // General purpose variable
  int16_t adc0, adc1, adc2, adc3, adc4, adc5, adc6, adc7;                 // Holds RAW Analog-to-Digital results
  float volts0, volts1, volts2, volts3, volts4, volts5, volts6, volts7;   // Holds DAC to Volts conversion results

  
  // Generate pulses on designated MCP23017 port pin
  for (i=0; i < PORT_PULSCNT; i++)
  {
    MCP23017.digitalWrite(PORT_PIN, HIGH);
    MCP23017.digitalWrite(PORT_PIN, LOW);
  }

  /* * * * * * * * * * * * * * * * * * * *
   * Write Values / Set Output Voltages on the DAC chips
   */
  // Write AOCH[3:0] controlled by MCP4728 DAC U2
  MCP4728_A.setChannelValue(MCP4728_CHANNEL_A, 4000);       // AOCH-0
  MCP4728_A.setChannelValue(MCP4728_CHANNEL_B, 2048);       // AOCH-1
  MCP4728_A.setChannelValue(MCP4728_CHANNEL_C, 1024);       // AOCH-2
  MCP4728_A.setChannelValue(MCP4728_CHANNEL_D, 0);          // AOCH-3

  // Write AOCH[7:4] controlled by MCP4728 DAC U3
  MCP4728_B.setChannelValue(MCP4728_CHANNEL_A, 400);        // AOCH-4
  MCP4728_B.setChannelValue(MCP4728_CHANNEL_B, 300);        // AOCH-5
  MCP4728_B.setChannelValue(MCP4728_CHANNEL_C, 200);        // AOCH-6
  MCP4728_B.setChannelValue(MCP4728_CHANNEL_D, 100);        // AOCH-7


  /* * * * * * * * * * * * * * * * * * * *
   * Digitize ADC Channels 0 - 3, ADS1115 U4
   * Calculate the voltage and display on Arduino IDE
   */
  adc0 = ADS1115_A.readADC_SingleEnded(0);                  // AICH-0
  adc1 = ADS1115_A.readADC_SingleEnded(1);                  // AICH-1
  adc2 = ADS1115_A.readADC_SingleEnded(2);                  // AICH-2
  adc3 = ADS1115_A.readADC_SingleEnded(3);                  // AICH-3

  volts0 = ADS1115_A.computeVolts(adc0);                    // AICH-0
  volts1 = ADS1115_A.computeVolts(adc1);                    // AICH-1
  volts2 = ADS1115_A.computeVolts(adc2);                    // AICH-2
  volts3 = ADS1115_A.computeVolts(adc3);                    // AICH-3

  Serial.println(F("-----------------"));
  Serial.print(F("AIN0: ")); Serial.print(adc0); Serial.print(F("  ")); Serial.print(volts0); Serial.println(F("V"));
  Serial.print(F("AIN1: ")); Serial.print(adc1); Serial.print(F("  ")); Serial.print(volts1); Serial.println(F("V"));
  Serial.print(F("AIN2: ")); Serial.print(adc2); Serial.print(F("  ")); Serial.print(volts2); Serial.println(F("V"));
  Serial.print(F("AIN3: ")); Serial.print(adc3); Serial.print(F("  ")); Serial.print(volts3); Serial.println(F("V"));


  /* * * * * * * * * * * * * * * * * * * *
   * Digitize ADC Channels 4 - 7, ADS1115 U5
   * Calculate the voltage and display on Arduino IDE
   */

  adc4 = ADS1115_B.readADC_SingleEnded(0);
  adc5 = ADS1115_B.readADC_SingleEnded(1);
  adc6 = ADS1115_B.readADC_SingleEnded(2);
  adc7 = ADS1115_B.readADC_SingleEnded(3);

  volts4 = ADS1115_B.computeVolts(adc4);
  volts5 = ADS1115_B.computeVolts(adc5);
  volts6 = ADS1115_B.computeVolts(adc6);
  volts7 = ADS1115_B.computeVolts(adc7);

  Serial.print(F("AIN4: ")); Serial.print(adc4); Serial.print(F("  ")); Serial.print(volts4); Serial.println(F("V"));
  Serial.print(F("AIN5: ")); Serial.print(adc5); Serial.print(F("  ")); Serial.print(volts5); Serial.println(F("V"));
  Serial.print(F("AIN6: ")); Serial.print(adc6); Serial.print(F("  ")); Serial.print(volts6); Serial.println(F("V"));
  Serial.print(F("AIN7: ")); Serial.print(adc7); Serial.print(F("  ")); Serial.print(volts7); Serial.println(F("V"));


  delay(1000);   // Pause loop for 1-Sec so user can observe screen

}
